import { Entry } from "./entry.js";
import { deserialize } from "@dao-xyz/borsh";
import { logger } from "./logger.js";
export class EntryIndex {
    _cache;
    _store;
    _init;
    _index;
    constructor(properties) {
        this._cache = properties.cache;
        this._store = properties.store;
        this._init = properties.init;
        this._index = new Set();
    }
    async set(v, toMultihash = true) {
        if (toMultihash) {
            const existingHash = v.hash;
            v.hash = undefined;
            try {
                const hash = await Entry.toMultihash(this._store, v);
                v.hash = existingHash;
                if (v.hash === undefined) {
                    v.hash = hash; // can happen if you sync entries that you load directly from ipfs
                }
                else if (existingHash !== v.hash) {
                    logger.error("Head hash didn't match the contents");
                    throw new Error("Head hash didn't match the contents");
                }
            }
            catch (error) {
                logger.error(error);
                throw error;
            }
        }
        this._cache.add(v.hash, v);
        this._index.add(v.hash);
    }
    async get(k, options) {
        if (this._index.has(k)) {
            let mem = this._cache.get(k);
            if (mem === undefined) {
                mem = await this.getFromStore(k, options);
                if (mem) {
                    this._init(mem);
                    mem.hash = k;
                }
                this._cache.add(k, mem);
            }
            return mem ? mem : undefined;
        }
        return undefined;
    }
    async getFromStore(k, options) {
        const value = await this._store.get(k, options);
        return value ? deserialize(value.value, Entry) : null;
    }
    async delete(k) {
        this._cache.del(k);
        this._index.delete(k);
        return this._store.rm(k);
    }
}
//# sourceMappingURL=entry-index.js.map