var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
import PQueue from "p-queue";
import { v4 as uuid } from "uuid";
import { sha256Base64Sync } from "@dao-xyz/peerbit-crypto";
import { variant, option, field, vec } from "@dao-xyz/borsh";
import { serialize, deserialize } from "@dao-xyz/borsh";
import { logger as loggerFn } from "@dao-xyz/peerbit-logger";
import path from "path-browserify";
export const logger = loggerFn({ module: "heads-cache" });
export class CachedValue {
}
/* export type AppendOptions<T> = {
    signers?: ((data: Uint8Array) => Promise<SignatureWithKey>)[];
    nexts?: Entry<T>[];
    reciever?: EncryptionTemplateMaybeEncrypted;
    type?: EntryType;
};
 */
let CachePath = class CachePath {
    path;
    constructor(path) {
        this.path = path;
    }
};
__decorate([
    field({ type: "string" }),
    __metadata("design:type", String)
], CachePath.prototype, "path", void 0);
CachePath = __decorate([
    variant(0),
    __metadata("design:paramtypes", [String])
], CachePath);
export { CachePath };
let UnsfinishedReplication = class UnsfinishedReplication {
    hashes;
    constructor(opts) {
        if (opts) {
            this.hashes = opts.hashes;
        }
    }
};
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], UnsfinishedReplication.prototype, "hashes", void 0);
UnsfinishedReplication = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], UnsfinishedReplication);
export { UnsfinishedReplication };
let HeadsCacheToSerialize = class HeadsCacheToSerialize {
    heads;
    last;
    counter;
    constructor(heads, counter, last) {
        this.heads = heads;
        this.last = last;
        this.counter = counter;
    }
};
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], HeadsCacheToSerialize.prototype, "heads", void 0);
__decorate([
    field({ type: option("string") }),
    __metadata("design:type", String)
], HeadsCacheToSerialize.prototype, "last", void 0);
__decorate([
    field({ type: "u64" }),
    __metadata("design:type", BigInt)
], HeadsCacheToSerialize.prototype, "counter", void 0);
HeadsCacheToSerialize = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Array, BigInt, String])
], HeadsCacheToSerialize);
export { HeadsCacheToSerialize };
export class HeadsCache {
    // An access controller that is note part of the store manifest, usefull for circular store -> access controller -> store structures
    headsPath;
    removedHeadsPath;
    initialized;
    _headsPathCounter = 0;
    _lastHeadsPath;
    _lastHeadsCount = 0n;
    _lastRemovedHeadsPath;
    _lastRemovedHeadsCount = 0n;
    _cache;
    _cacheWriteQueue;
    _loaded = false;
    _index;
    constructor(index) {
        this._index = index;
    }
    async init(cache) {
        if (this.initialized) {
            throw new Error("Already initialized");
        }
        this._cache = await cache(sha256Base64Sync(this._index.id));
        // Set the options (we will use the topic property after thiis)
        await this._cache.open();
        this.headsPath = "heads";
        this.removedHeadsPath = "heads_removed";
        await this.loadLastHeadsPath();
        // append and log-joins queue. Adding ops and joins to the queue
        // makes sure they get processed sequentially to avoid race conditions
        this._cacheWriteQueue = new PQueue({ concurrency: 1 });
        /* if (this._options.onOpen) {
            await this._options.onOpen();
        } */
        this.initialized = true;
        return this;
    }
    get loaded() {
        return this._loaded;
    }
    async _updateCachedHeads(change, reset) {
        if (typeof reset !== "boolean" && change.added) {
            // Only reset all heads if loaded once, since we don't want too loose track of unloaded heads
            if (this._loaded && this._index.size <= change.added.length) {
                let addedIsAllHeads = true;
                for (const entry of change.added) {
                    const hash = typeof entry === "string" ? entry : entry.hash;
                    if (!this._index.has(hash)) {
                        addedIsAllHeads = false;
                    }
                }
                reset = addedIsAllHeads;
            }
            else {
                // added size < head size, meaning we have not rewritten all heads
                reset = false;
            }
        }
        // If 'reset' then dont keep references to old heads caches, assume new cache will fully describe all heads
        const updateHashes = async (headsPath, lastCid, lastCounter, hashes) => {
            const newHeadsPath = path.join(headsPath, String(this._headsPathCounter), uuid());
            const counter = lastCounter + BigInt(hashes.length);
            await Promise.all([
                this._cache.set(headsPath, serialize(new CachePath(newHeadsPath.toString()))),
                this._cache.set(newHeadsPath, serialize(new HeadsCacheToSerialize(hashes, counter, lastCid))),
            ]);
            return { counter, newPath: newHeadsPath };
        };
        // TODO dont delete old before saving new
        if (reset) {
            const paths = [
                path.join(this.headsPath, String(this._headsPathCounter)),
                path.join(this.removedHeadsPath, String(this._headsPathCounter)),
            ];
            for (const p of paths) {
                await this._cache.deleteByPrefix(p + "/");
            }
            this._lastHeadsPath = undefined;
            this._lastRemovedHeadsPath = undefined;
            this._lastHeadsCount = 0n;
            this._lastRemovedHeadsCount = 0n;
            this._headsPathCounter += 1;
        }
        if (change.added && change.added.length > 0) {
            const update = await updateHashes(this.headsPath, this._lastHeadsPath, this._lastHeadsCount, change.added.map((x) => (typeof x === "string" ? x : x.hash)));
            this._lastHeadsPath = update.newPath;
            this._lastHeadsCount = update.counter;
        }
        if (this._lastHeadsPath) {
            // only add removed heads if we actually have added heads, else these are pointless
            if (change.removed && change.removed.length > 0) {
                const update = await updateHashes(this.removedHeadsPath, this._lastRemovedHeadsPath, this._lastRemovedHeadsCount, change.removed.map((x) => (typeof x === "string" ? x : x.hash)));
                this._lastRemovedHeadsPath = update.newPath;
                this._lastRemovedHeadsCount = update.counter;
                if (update.counter > 0n &&
                    2n * update.counter >= this._lastHeadsCount) {
                    const resetToHeads = await this.getCachedHeads(this._lastHeadsPath, this._lastRemovedHeadsPath);
                    await this._updateCachedHeads({ added: resetToHeads, removed: [] }, true);
                }
            }
        }
    }
    async idle() {
        // Wait for the operations queue to finish processing
        // to make sure everything that all operations that have
        // been queued will be written to disk
        await this._cacheWriteQueue?.onIdle();
        await this._cache?.idle();
    }
    async getCachedHeads(lastHeadsPath = this._lastHeadsPath, lastRemovedHeadsPath = this._lastRemovedHeadsPath) {
        if (!this._cache) {
            return [];
        }
        const getHashes = async (start, filter) => {
            const result = [];
            let next = start;
            while (next) {
                const cache = await this._cache
                    .get(next)
                    .then((bytes) => bytes && deserialize(bytes, HeadsCacheToSerialize));
                next = cache?.last;
                cache?.heads.forEach((head) => {
                    if (filter && filter.has(head)) {
                        return;
                    }
                    result.push(head);
                });
            }
            return result;
        };
        const removedHeads = new Set(await getHashes(lastRemovedHeadsPath));
        const heads = await getHashes(lastHeadsPath, removedHeads);
        return heads; // Saved heads - removed heads
    }
    /* get logOptions(): LogOptions<T> {
        return {
            logId: this.id,
            trim: this._options.trim && {
                // I can trim if I am not a replicator of an entry

                ...this._options.trim,
                filter: this.options.replicator && {
                    canTrim: async (gid) => !(await this.options.replicator!(gid)),
                    cacheId: this.options.replicatorsCacheId,
                },
            },
        };
    } */
    get cache() {
        return this._cache;
    }
    get closed() {
        return this._cache.status === "closed";
    }
    async close() {
        if (!this.initialized) {
            return;
        }
        await this.idle();
        await this._cache.close();
        this._loaded = false;
        this._lastHeadsPath = undefined;
        this._lastRemovedHeadsPath = undefined;
        this._lastRemovedHeadsCount = 0n;
        this._lastHeadsCount = 0n;
        // Database is now closed
        return Promise.resolve();
    }
    /**
     * Drops a database and removes local data
     */
    async drop() {
        if (!this._cache) {
            return; // already dropped
        }
        if (this._cache.status !== "open") {
            await this._cache.open();
        }
        await this._cache.del(this.headsPath);
        await this._cache.del(this.removedHeadsPath);
        await this.close();
        // Reset
        // TODO fix types
        this._cache = undefined;
        this.initialized = false; // call this last because (close() expect initialized to be able to function)
    }
    async _loadHeads() {
        if (!this.initialized) {
            throw new Error("Store needs to be initialized before loaded");
        }
        if (this._cache.status !== "open") {
            await this._cache.open();
        }
        await this.loadLastHeadsPath();
        return this.getCachedHeads();
    }
    async load() {
        if (!this.initialized) {
            throw new Error("Needs to be initialized before loaded");
        }
        if (this._cache.status !== "open") {
            await this._cache.open();
        }
        const heads = await this._loadHeads();
        this._loaded = true;
        return heads;
    }
    async loadLastHeadsPath() {
        this._lastHeadsPath = await this._cache
            .get(this.headsPath)
            .then((bytes) => bytes && deserialize(bytes, CachePath).path);
        this._lastRemovedHeadsPath = await this._cache
            .get(this.removedHeadsPath)
            .then((bytes) => bytes && deserialize(bytes, CachePath).path);
        this._lastHeadsCount = this._lastHeadsPath
            ? await this.getCachedHeadsCount(this._lastHeadsPath)
            : 0n;
        this._lastRemovedHeadsCount = this._lastRemovedHeadsPath
            ? await this.getCachedHeadsCount(this._lastRemovedHeadsPath)
            : 0n;
    }
    async getCachedHeadsCount(headPath) {
        if (!headPath) {
            return 0n;
        }
        return ((await this._cache
            .get(headPath)
            .then((bytes) => bytes && deserialize(bytes, HeadsCacheToSerialize)))?.counter || 0n);
    }
    async waitForHeads() {
        if (this.closed) {
            throw new Error("Store is closed");
        }
        if (!this._loaded) {
            return this._cacheWriteQueue.add(async () => {
                if (this._loaded) {
                    return;
                }
                return this.load();
            });
        }
    }
    queue(changes, reset) {
        return this._cacheWriteQueue.add(() => this._updateCachedHeads(changes, reset));
    }
}
//# sourceMappingURL=heads-cache.js.map