import { Entry } from "./entry.js";
import LocalStore from "@dao-xyz/lazy-level";
import { HeadsCache } from "./heads-cache.js";
import { BlockStore } from "@dao-xyz/libp2p-direct-block";
export type CacheUpdateOptions = {
    cache?: {
        update?: false;
        reset?: false;
    } | {
        update: true;
        reset?: boolean;
    };
};
export declare class HeadsIndex<T> {
    private _id;
    private _index;
    private _gids;
    private _headsCache;
    private _blockstore;
    constructor(id: Uint8Array);
    init(blockstore: BlockStore, cache?: (name: string) => Promise<LocalStore> | LocalStore, options?: {
        entries?: Entry<T>[];
    }): Promise<HeadsCache<T> | undefined>;
    load(options?: {
        timeout?: number;
        replicate?: boolean;
        reload?: boolean;
    } & CacheUpdateOptions): Promise<Entry<T>[] | undefined>;
    get headsCache(): HeadsCache<T> | undefined;
    close(): Promise<void> | undefined;
    drop(): Promise<void> | undefined;
    get id(): Uint8Array;
    get index(): Set<string>;
    get gids(): Map<string, number>;
    get size(): number;
    reset(entries: Entry<T>[], options?: CacheUpdateOptions): Promise<void>;
    has(cid: string): boolean;
    put(entry: Entry<T>, options?: CacheUpdateOptions): Promise<void>;
    putAll(entries: Entry<T>[], options?: CacheUpdateOptions): Promise<void>;
    resetHeadsCache(): Promise<void>;
    updateHeadsCache(change?: {
        added?: (Entry<T> | string)[];
        removed?: (Entry<T> | string)[];
    }, reset?: boolean): Promise<void>;
    private _putOne;
    private _putAll;
    del(entry: {
        hash: string;
        gid: string;
    }, options?: CacheUpdateOptions): Promise<{
        removed: boolean;
        lastWithGid: boolean;
    }>;
}
