var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var Log_1;
import { randomBytes, sha256Base64Sync, } from "@dao-xyz/peerbit-crypto";
import { Cache } from "@dao-xyz/cache";
import { EntryIndex } from "./entry-index.js";
import * as LogError from "./log-errors.js";
import * as Sorting from "./log-sorting.js";
import { isDefined } from "./is-defined.js";
import { findUniques } from "./find-uniques.js";
import { Entry, EntryType, } from "./entry.js";
import { HLC, LamportClock as Clock, } from "./clock.js";
import { field, fixedArray, variant } from "@dao-xyz/borsh";
import { JSON_ENCODING } from "./encoding.js";
import { HeadsIndex } from "./heads.js";
import { Values } from "./values.js";
import { Trim } from "./trim.js";
import { logger } from "./logger.js";
const { LastWriteWins, NoZeroes } = Sorting;
const ENTRY_CACHE_MAX = 1000; // TODO as param
let Log = Log_1 = class Log {
    _id;
    _sortFn;
    _storage;
    _hlc;
    // Identity
    _identity;
    // Keeping track of entries
    _entryIndex;
    _headsIndex;
    _values;
    // Index of all next pointers in this log
    _nextsIndex;
    _encryption;
    _encoding;
    _trim;
    _entryCache;
    _canAppendInit;
    _canAppend;
    _onChangeInit;
    _onChange;
    _onWrite;
    _onClose;
    _joining; // entry hashes that are currently joining into this log
    constructor(properties = { id: randomBytes(32) }) {
        this._id = properties.id;
    }
    async init(store, identity, options = {}) {
        if (!isDefined(store)) {
            throw LogError.BlockStoreNotDefinedError();
        }
        if (!isDefined(identity)) {
            throw new Error("Identity is required");
        }
        //
        const { encoding, trim, encryption, onChange, cache, onWrite, onClose } = options;
        let { sortFn } = options;
        this._onChangeInit = onChange;
        this._onWrite = onWrite;
        this._onClose = onClose;
        if (!isDefined(sortFn)) {
            sortFn = LastWriteWins;
        }
        sortFn = sortFn;
        this._sortFn = NoZeroes(sortFn);
        this._storage = store;
        this._encoding = encoding || JSON_ENCODING;
        this._joining = new Map();
        // Identity
        this._identity = identity;
        // encoder/decoder
        this._encryption = encryption;
        // Add entries to the internal cache
        // Index of all next pointers in this log
        // Clock
        this._hlc = new HLC();
        this._nextsIndex = new Map();
        this._headsIndex = new HeadsIndex(this._id);
        await this._headsIndex.init(store, cache);
        this._entryCache = new Cache({ max: ENTRY_CACHE_MAX });
        this._entryIndex = new EntryIndex({
            store: this._storage,
            init: (e) => e.init(this),
            cache: this._entryCache,
        });
        this._values = new Values(this._entryIndex, this._sortFn);
        this._trim = new Trim({
            deleteNode: async (node) => {
                // TODO check if we have before delete?
                const entry = await this.get(node.value.hash);
                //f (!!entry)
                const a = this.values.length;
                if (entry) {
                    this.values.deleteNode(node);
                    await this.entryIndex.delete(node.value.hash);
                    await this.headsIndex.del(node.value);
                    this.nextsIndex.delete(node.value.hash);
                    await this.storage.rm(node.value.hash);
                }
                const b = this.values.length;
                if (a === b) {
                    throw new Error("UNexpected: " +
                        this.values.length +
                        "_-- " +
                        this.entryIndex._index.size);
                }
                return entry;
            },
            values: () => this.values,
        }, trim);
    }
    async setup(options) {
        this._encoding = options?.encoding || this._encoding;
        this._canAppend = async (entry) => {
            if (this._canAppendInit) {
                if (!(await this._canAppendInit(entry))) {
                    return false;
                }
            }
            if (options?.canAppend) {
                if (!(await options.canAppend(entry))) {
                    return false;
                }
            }
            return true;
        };
        this._onChange = async (change) => {
            await this._onChangeInit?.(change);
            await options?.onChange?.(change);
        };
    }
    get initialized() {
        return !!this._storage;
    }
    _idString;
    get idString() {
        return this._idString || (this._idString = Log_1.createIdString(this.id));
    }
    static createIdString(id) {
        return sha256Base64Sync(id);
    }
    get id() {
        return this._id;
    }
    set id(id) {
        if (this.initialized) {
            throw new Error("Can not change id after initialization");
        }
        this._idString = undefined;
        this._id = id;
    }
    /* set canAppend(canAppend: CanAppend<T> | undefined) {
        this._canAppendOption = canAppend;
    }
    get canAppend(): CanAppend<T> | undefined {
        return this._canAppend;
    } */
    /**
     * Returns the length of the log.
     */
    get length() {
        return this._values.length;
    }
    get values() {
        return this._values;
    }
    /**
     * Checks if a entry is part of the log
     * @param {string} hash The hash of the entry
     * @returns {boolean}
     */
    has(cid) {
        return this._entryIndex._index.has(cid);
    }
    /**
     * Get all entries sorted. Don't use this method anywhere where performance matters
     */
    toArray() {
        // we call init, because the values might be unitialized
        return this._values.toArray().then((arr) => arr.map((x) => x.init(this)));
    }
    /**
     * Returns the head index
     */
    get headsIndex() {
        return this._headsIndex;
    }
    /**
     * Don't use this anywhere performance matters
     */
    async getHeads() {
        const heads = new Array(this.headsIndex.index.size);
        let i = 0;
        for (const hash of this.headsIndex.index) {
            heads[i++] = this._entryIndex.get(hash).then((x) => x?.init(this));
        }
        const resolved = await Promise.all(heads);
        const defined = resolved.filter((x) => !!x);
        if (defined.length !== resolved.length) {
            logger.error("Failed to resolve all heads");
        }
        return defined;
    }
    /**
     * Returns an array of Entry objects that reference entries which
     * are not in the log currently.
     * @returns {Array<Entry<T>>}
     */
    async getTails() {
        return Log_1.findTails(await this.toArray());
    }
    /**
     * Returns an array of hashes that are referenced by entries which
     * are not in the log currently.
     * @returns {Array<string>} Array of hashes
     */
    async getTailHashes() {
        return Log_1.findTailHashes(await this.toArray());
    }
    /**
     * Get local HLC
     */
    get hlc() {
        return this._hlc;
    }
    get identity() {
        return this._identity;
    }
    get storage() {
        return this._storage;
    }
    get nextsIndex() {
        return this._nextsIndex;
    }
    get entryIndex() {
        return this._entryIndex;
    }
    get encryption() {
        return this._encryption;
    }
    get encoding() {
        return this._encoding;
    }
    get sortFn() {
        return this._sortFn;
    }
    /**
     * Set the identity for the log
     * @param {Identity} [identity] The identity to be set
     */
    setIdentity(identity) {
        this._identity = identity;
    }
    /**
     * Find an entry.
     * @param {string} [hash] The hashes of the entry
     */
    get(hash, options) {
        return this._entryIndex.get(hash, options);
    }
    async traverse(rootEntries, amount = -1, endHash) {
        // Sort the given given root entries and use as the starting stack
        let stack = rootEntries.sort(this._sortFn).reverse();
        // Cache for checking if we've processed an entry already
        let traversed = {};
        // End result
        const result = {};
        let count = 0;
        // Named function for getting an entry from the log
        const getEntry = (e) => this.get(e);
        // Add an entry to the stack and traversed nodes index
        const addToStack = (entry) => {
            // If we've already processed the Entry<T>, don't add it to the stack
            if (!entry || traversed[entry.hash]) {
                return;
            }
            // Add the entry in front of the stack and sort
            stack = [entry, ...stack].sort(this._sortFn).reverse();
            // Add to the cache of processed entries
            traversed[entry.hash] = true;
        };
        const addEntry = (rootEntry) => {
            result[rootEntry.hash] = rootEntry;
            traversed[rootEntry.hash] = true;
            count++;
        };
        // Start traversal
        // Process stack until it's empty (traversed the full log)
        // or when we have the requested amount of entries
        // If requested entry amount is -1, traverse all
        while (stack.length > 0 && (count < amount || amount < 0)) {
            // eslint-disable-line no-unmodified-loop-condition
            // Get the next element from the stack
            const entry = stack.shift();
            if (!entry) {
                throw new Error("Unexpected");
            }
            // Add to the result
            addEntry(entry);
            // If it is the specified end hash, break out of the while loop
            if (endHash && endHash === entry.hash)
                break;
            // Add entry's next references to the stack
            const entries = (await Promise.all(entry.next.map(getEntry))).filter((x) => !!x);
            entries.forEach(addToStack);
        }
        stack = [];
        traversed = {};
        // End result
        return result;
    }
    async getReferenceSamples(from, options) {
        const hashes = new Set();
        const pointerCount = options?.pointerCount || 0;
        const memoryLimit = options?.memoryLimit;
        const maxDistance = Math.min(pointerCount, this._values.length);
        if (maxDistance === 0) {
            return [];
        }
        hashes.add(from.hash);
        let memoryCounter = from._payload.byteLength;
        if (from.next?.length > 0 && pointerCount >= 2) {
            let next = new Set(from.next);
            let prev = 2;
            outer: for (let i = 2; i <= maxDistance - 1; i *= 2) {
                for (let j = prev; j < i; j++) {
                    if (next.size === 0) {
                        break outer;
                    }
                    const nextNext = new Set();
                    for (const n of next) {
                        const nentry = await this.get(n);
                        nentry?.next.forEach((n2) => {
                            nextNext.add(n2);
                        });
                    }
                    next = nextNext;
                }
                prev = i;
                if (next) {
                    for (const n of next) {
                        if (!memoryLimit) {
                            hashes.add(n);
                        }
                        else {
                            const entry = await this.get(n);
                            if (!entry) {
                                break outer;
                            }
                            memoryCounter += entry._payload.byteLength;
                            if (memoryCounter > memoryLimit) {
                                break outer;
                            }
                            hashes.add(n);
                        }
                        if (hashes.size === pointerCount) {
                            break outer;
                        }
                    }
                }
            }
        }
        const ret = [];
        for (const hash of hashes) {
            const entry = await this.get(hash);
            if (entry) {
                ret.push(entry);
            }
        }
        return ret;
    }
    /**
     * Append an entry to the log.
     * @param {Entry} entry Entry to add
     * @return {Log} New Log containing the appended value
     */
    async append(data, options = {}) {
        if (options.reciever && !this._encryption) {
            throw new Error("Message is intended to be encrypted but no encryption methods are provided for the log");
        }
        // Update the clock (find the latest clock)
        if (options.nexts) {
            for (const n of options.nexts) {
                if (!n.hash)
                    throw new Error("Expecting nexts to already be saved. missing hash for one or more entries");
            }
        }
        await this.load({ reload: false });
        const hasNext = !!options.nexts; // true for [], which means we have explicitly said that nexts are empty
        const nexts = options.nexts || (await this.getHeads());
        // Calculate max time for log/graph
        const clock = new Clock({
            id: this._identity.publicKey.bytes,
            timestamp: options.timestamp || this._hlc.now(),
        });
        const entry = await Entry.create({
            store: this._storage,
            identity: options.identity || this._identity,
            signers: options.signers,
            data,
            clock,
            type: options.type,
            encoding: this._encoding,
            next: nexts,
            gidSeed: options.gidSeed,
            encryption: options.reciever
                ? {
                    options: this._encryption,
                    reciever: {
                        ...options.reciever,
                    },
                }
                : undefined,
            canAppend: this._canAppend,
        });
        if (!isDefined(entry.hash)) {
            throw new Error("Unexpected");
        }
        for (const e of nexts) {
            let nextIndexSet = this._nextsIndex.get(e.hash);
            if (!nextIndexSet) {
                nextIndexSet = new Set();
                nextIndexSet.add(entry.hash);
                this._nextsIndex.set(e.hash, nextIndexSet);
            }
            else {
                nextIndexSet.add(entry.hash);
            }
        }
        const removedGids = new Set();
        if (hasNext) {
            for (const next of nexts) {
                const deletion = await this._headsIndex.del(next);
                if (deletion.lastWithGid && next.gid !== entry.gid) {
                    removedGids.add(next.gid);
                }
            }
        }
        else {
            // next is all heads, which means we should just overwrite
            for (const key of this.headsIndex.gids.keys()) {
                if (key !== entry.gid) {
                    removedGids.add(key);
                }
            }
            await this.headsIndex.reset([entry], { cache: { update: false } });
        }
        await this._entryIndex.set(entry, false); // save === false, because its already saved when Entry.create
        await this._headsIndex.put(entry, { cache: { update: false } }); // we will update the cache a few lines later *
        await this._values.put(entry);
        const removed = await this.processEntry(entry);
        // if next contails all gids
        if (options.onGidsShadowed && removedGids.size > 0) {
            options.onGidsShadowed([...removedGids]);
        }
        entry.init({ encoding: this._encoding, encryption: this._encryption });
        //	console.log('put entry', entry.hash, (await this._entryIndex._index.size));
        const trimmed = await this.trim(options?.trim);
        for (const entry of trimmed) {
            removed.push(entry);
        }
        const changes = {
            added: [entry],
            removed: removed,
        };
        await this._headsIndex.updateHeadsCache(changes); // * here
        await this._onWrite?.(entry);
        await this._onChange?.(changes);
        return { entry, removed };
    }
    async reset(entries, heads) {
        this._nextsIndex = new Map();
        this._entryIndex = new EntryIndex({
            store: this._storage,
            init: (e) => e.init(this),
            cache: this._entryCache,
        });
        const promises = [];
        const set = new Set();
        const uniqueEntries = [];
        for (const entry of entries) {
            if (!entry.hash) {
                throw new Error("Unexpected");
            }
            if (set.has(entry.hash)) {
                continue;
            }
            set.add(entry.hash);
            uniqueEntries.push(entry);
            promises.push(this._entryIndex.set(entry));
        }
        await Promise.all(promises);
        // Set heads if not passed as an argument
        const foundHeads = heads
            ? (await Promise.all(heads.map((x) => {
                if (x instanceof Entry)
                    return x;
                const resolved = this._entryIndex.get(x);
                if (!resolved) {
                    throw new Error("Missing head with cid: " + x);
                }
                return resolved;
            })))
            : Log_1.findHeads(uniqueEntries);
        await this._headsIndex.reset(foundHeads);
        this._values = new Values(this._entryIndex, this._sortFn, uniqueEntries);
        for (const e of entries) {
            for (const a of e.next) {
                let nextIndexSet = this._nextsIndex.get(a);
                if (!nextIndexSet) {
                    nextIndexSet = new Set();
                    nextIndexSet.add(e.hash);
                    this._nextsIndex.set(a, nextIndexSet);
                }
                else {
                    nextIndexSet.add(e.hash);
                }
            }
        }
    }
    async remove(entry, options) {
        await this.load({ reload: false });
        const entries = Array.isArray(entry) ? entry : [entry];
        if (entries.length === 0) {
            return {
                added: [],
                removed: [],
            };
        }
        if (options?.recursively) {
            await this.deleteRecursively(entry);
        }
        else {
            for (const entry of entries) {
                await this.delete(entry);
            }
        }
        const change = {
            added: [],
            removed: Array.isArray(entry) ? entry : [entry],
        };
        /* 	await Promise.all([
                this._logCache?.queue(change),
                this._onUpdate(change),
            ]); */
        await this._onChange?.(change);
        return change;
    }
    iterator(options) {
        const from = options?.from || "tail";
        const amount = typeof options?.amount === "number" ? options?.amount : -1;
        let next = from === "tail" ? this._values.tail : this._values.head;
        const nextFn = from === "tail" ? (e) => e.prev : (e) => e.next;
        return (function* () {
            let counter = 0;
            while (next) {
                if (amount >= 0 && counter >= amount) {
                    return;
                }
                yield next.value.hash;
                counter++;
                next = nextFn(next);
            }
        })();
    }
    async trim(option = this._trim.options) {
        return this._trim.trim(option);
    }
    /**
     *
     * @param entries
     * @returns change
     */
    /* async sync(
        entries: (EntryWithRefs<T> | Entry<T> | string)[],
        options: {
            canAppend?: CanAppend<T>;
            onChange?: (change: Change<T>) => void | Promise<void>;
            timeout?: number;
        } = {}
    ): Promise<void> {


        logger.debug(`Sync request #${entries.length}`);
        const entriesToJoin: (Entry<T> | string)[] = [];
        for (const e of entries) {
            if (e instanceof Entry || typeof e === "string") {
                entriesToJoin.push(e);
            } else {
                for (const ref of e.references) {
                    entriesToJoin.push(ref);
                }
                entriesToJoin.push(e.entry);
            }
        }

        await this.join(entriesToJoin, {
            canAppend: (entry) => {
                const canAppend = options?.canAppend || this.canAppend;
                return !canAppend || canAppend(entry);
            },
            onChange: (change) => {
                options?.onChange?.(change);
                return this._onChange?.({
                    added: change.added,
                    removed: change.removed,
                });
            },
            timeout: options.timeout,
        });
    } */
    async join(entriesOrLog, options) {
        await this.load({ reload: false });
        if (entriesOrLog.length === 0) {
            return;
        }
        /* const joinLength = options?.length ?? Number.MAX_SAFE_INTEGER;  TODO */
        const visited = new Set();
        const nextRefs = new Map();
        const entriesBottomUp = [];
        const stack = [];
        const resolvedEntries = new Map();
        const entries = Array.isArray(entriesOrLog)
            ? entriesOrLog
            : await entriesOrLog.values.toArray();
        // Build a list of already resolved entries, and filter out already joined entries
        for (const e of entries) {
            // TODO, do this less ugly
            let hash;
            if (e instanceof Entry) {
                hash = e.hash;
                resolvedEntries.set(e.hash, e);
                if (this.has(hash)) {
                    continue;
                }
                stack.push(hash);
            }
            else if (typeof e === "string") {
                hash = e;
                if (this.has(hash)) {
                    continue;
                }
                stack.push(hash);
            }
            else {
                hash = e.entry.hash;
                resolvedEntries.set(e.entry.hash, e.entry);
                if (this.has(hash)) {
                    continue;
                }
                stack.push(hash);
                for (const e2 of e.references) {
                    resolvedEntries.set(e2.hash, e2);
                    if (this.has(e2.hash)) {
                        continue;
                    }
                    stack.push(e2.hash);
                }
            }
        }
        // Resolve missing entries
        const removedHeads = [];
        for (const hash of stack) {
            if (visited.has(hash) || this.has(hash)) {
                continue;
            }
            visited.add(hash);
            const entry = resolvedEntries.get(hash) ||
                (await Entry.fromMultihash(this._storage, hash, {
                    timeout: options?.timeout,
                }));
            entry.init(this);
            resolvedEntries.set(entry.hash, entry);
            let nexts;
            if (entry.metadata.type !== EntryType.CUT &&
                (nexts = await entry.getNext())) {
                let isRoot = true;
                for (const next of nexts) {
                    if (!this.has(next)) {
                        isRoot = false;
                    }
                    else {
                        if (this._headsIndex.has(next)) {
                            const toRemove = (await this.get(next, options));
                            await this._headsIndex.del(toRemove);
                            removedHeads.push(toRemove);
                        }
                    }
                    let nextIndexSet = nextRefs.get(next);
                    if (!nextIndexSet) {
                        nextIndexSet = [];
                        nextIndexSet.push(entry);
                        nextRefs.set(next, nextIndexSet);
                    }
                    else {
                        nextIndexSet.push(entry);
                    }
                    if (!visited.has(next)) {
                        stack.push(next);
                    }
                }
                if (isRoot) {
                    entriesBottomUp.push(entry);
                }
            }
            else {
                entriesBottomUp.push(entry);
            }
        }
        while (entriesBottomUp.length > 0) {
            const e = entriesBottomUp.shift();
            await this._joining.get(e.hash);
            const p = this.joinEntry(e, nextRefs, entriesBottomUp, options).then(() => this._joining.delete(e.hash) // TODO, if head we run into problems with concurrency here!, we add heads at line 929 but resolve here
            );
            this._joining.set(e.hash, p);
            await p;
        }
    }
    async joinEntry(e, nextRefs, stack, options) {
        if (this.length > (options?.length ?? Number.MAX_SAFE_INTEGER)) {
            return;
        }
        if (!isDefined(e.hash)) {
            throw new Error("Unexpected");
        }
        if (!this.has(e.hash)) {
            if (options?.verifySignatures) {
                if (!(await e.verifySignatures())) {
                    throw new Error('Invalid signature entry with hash "' + e.hash + '"');
                }
            }
            if (this?._canAppend && !(await this?._canAppend(e))) {
                return;
            }
            // Update the internal entry index
            await this._entryIndex.set(e);
            await this._values.put(e);
            if (e.metadata.type !== EntryType.CUT) {
                for (const a of e.next) {
                    if (!this.has(a)) {
                        await this.join([a]);
                    }
                    let nextIndexSet = this._nextsIndex.get(a);
                    if (!nextIndexSet) {
                        nextIndexSet = new Set();
                        nextIndexSet.add(e.hash);
                        this._nextsIndex.set(a, nextIndexSet);
                    }
                    else {
                        nextIndexSet.add(a);
                    }
                }
            }
            const clock = await e.getClock();
            this._hlc.update(clock.timestamp);
            const removed = await this.processEntry(e);
            const trimmed = await this.trim(options?.trim);
            for (const entry of trimmed) {
                removed.push(entry);
            }
            await this?._onChange?.({ added: [e], removed: removed });
        }
        const forward = nextRefs.get(e.hash);
        if (forward) {
            if (this._headsIndex.has(e.hash)) {
                await this._headsIndex.del(e, options);
            }
            for (const en of forward) {
                stack.push(en);
            }
        }
        else {
            await this.headsIndex.put(e, options);
        }
    }
    async processEntry(entry) {
        if (entry.metadata.type === EntryType.CUT) {
            return this.deleteRecursively(entry, true);
        }
        return [];
    }
    /// TODO simplify methods below
    async deleteRecursively(from, skipFirst = false) {
        const stack = Array.isArray(from) ? [...from] : [from];
        const promises = [];
        let counter = 0;
        const deleted = [];
        while (stack.length > 0) {
            const entry = stack.pop();
            if ((counter > 0 || !skipFirst) && this.has(entry.hash)) {
                // TODO test last argument: It is for when multiple heads point to the same entry, hence we might visit it multiple times? or a concurrent delete process is doing it before us.
                this._trim.deleteFromCache(entry);
                await this._values.delete(entry);
                await this._entryIndex.delete(entry.hash);
                await this._headsIndex.del(entry);
                this._nextsIndex.delete(entry.hash);
                deleted.push(entry);
                promises.push(entry.delete(this._storage));
            }
            for (const next of entry.next) {
                const nextFromNext = this._nextsIndex.get(next);
                if (nextFromNext) {
                    nextFromNext.delete(entry.hash);
                }
                if (!nextFromNext || nextFromNext.size === 0) {
                    const ne = await this.get(next);
                    if (ne) {
                        stack.push(ne);
                    }
                }
            }
            counter++;
        }
        await Promise.all(promises);
        return deleted;
    }
    async delete(entry) {
        this._trim.deleteFromCache(entry);
        await this._values.delete(entry);
        await this._entryIndex.delete(entry.hash);
        await this._headsIndex.del(entry);
        this._nextsIndex.delete(entry.hash);
        const newHeads = [];
        for (const next of entry.next) {
            const ne = await this.get(next);
            if (ne) {
                const nexts = this._nextsIndex.get(next);
                nexts.delete(entry.hash);
                if (nexts.size === 0) {
                    await this._headsIndex.put(ne);
                    newHeads.push(ne.hash);
                }
            }
        }
        await this._headsIndex.updateHeadsCache({
            added: newHeads,
            removed: [entry.hash],
        });
        return entry.delete(this._storage);
    }
    /**
     * Returns the log entries as a formatted string.
     * @returns {string}
     * @example
     * two
     * └─one
     *   └─three
     */
    async toString(payloadMapper = (payload) => payload.getValue().toString()) {
        return (await Promise.all((await this.toArray())
            .slice()
            .reverse()
            .map(async (e, idx) => {
            const parents = Entry.findDirectChildren(e, await this.toArray());
            const len = parents.length;
            let padding = new Array(Math.max(len - 1, 0));
            padding = len > 1 ? padding.fill("  ") : padding;
            padding = len > 0 ? padding.concat(["└─"]) : padding;
            /* istanbul ignore next */
            return (padding.join("") +
                (payloadMapper ? payloadMapper(e.payload) : e.payload));
        }))).join("\n");
    }
    async idle() {
        await this._headsIndex.headsCache?.idle();
    }
    async close() {
        await this._entryCache.clear();
        await this._headsIndex.close();
        await this._onClose?.();
    }
    async drop() {
        await this.deleteRecursively(await this.getHeads()); // TODO can multiple store have exact same log entry? no because GIDs are generated randomly
        await this._headsIndex.drop();
        await this._entryCache.clear();
        return;
    }
    async load(opts = { reload: true }) {
        const heads = await this.headsIndex.load({
            replicate: true,
            timeout: opts.fetchEntryTimeout,
            reload: opts.reload,
            cache: { update: true, reset: true },
        });
        if (heads) {
            // Load the log
            if (opts.heads) {
                await this.reset(heads);
            }
            else {
                const amount = opts.amount;
                if (amount != null && amount >= 0 && amount < heads.length) {
                    throw new Error("You are not loading all heads, this will lead to unexpected behaviours on write. Please load at least load: " +
                        amount +
                        " entries");
                }
                await this.join(heads instanceof Entry ? [heads] : heads, {
                    /* length: amount, */
                    timeout: opts?.fetchEntryTimeout,
                    cache: {
                        update: false,
                    },
                });
            }
        }
    }
    static async fromEntry(store, identity, entryOrHash, options = { id: randomBytes(32) }) {
        const log = new Log_1(options.id && { id: options.id });
        await log.init(store, identity, options);
        await log.join(!Array.isArray(entryOrHash) ? [entryOrHash] : entryOrHash, {
            timeout: options.timeout,
            trim: options.trim,
            verifySignatures: true,
        });
        return log;
    }
    /**
     * Find heads from a collection of entries.
     *
     * Finds entries that are the heads of this collection,
     * ie. entries that are not referenced by other entries.
     *
     * @param {Array<Entry<T>>} entries Entries to search heads from
     * @returns {Array<Entry<T>>}
     */
    static findHeads(entries) {
        const indexReducer = (res, entry) => {
            const addToResult = (e) => (res[e] = entry.hash);
            entry.next.forEach(addToResult);
            return res;
        };
        const items = entries.reduce(indexReducer, {});
        const exists = (e) => items[e.hash] === undefined;
        return entries.filter(exists);
    }
    // Find entries that point to another entry that is not in the
    // input array
    static findTails(entries) {
        // Reverse index { next -> entry }
        const reverseIndex = {};
        // Null index containing entries that have no parents (nexts)
        const nullIndex = [];
        // Hashes for all entries for quick lookups
        const hashes = {};
        // Hashes of all next entries
        let nexts = [];
        const addToIndex = (e) => {
            if (e.next.length === 0) {
                nullIndex.push(e);
            }
            const addToReverseIndex = (a) => {
                /* istanbul ignore else */
                if (!reverseIndex[a])
                    reverseIndex[a] = [];
                reverseIndex[a].push(e);
            };
            // Add all entries and their parents to the reverse index
            e.next.forEach(addToReverseIndex);
            // Get all next references
            nexts = nexts.concat(e.next);
            // Get the hashes of input entries
            hashes[e.hash] = true;
        };
        // Create our indices
        entries.forEach(addToIndex);
        const addUniques = (res, entries, _idx, _arr) => res.concat(findUniques(entries, "hash"));
        const exists = (e) => hashes[e] === undefined;
        const findFromReverseIndex = (e) => reverseIndex[e];
        // Drop hashes that are not in the input entries
        const tails = nexts // For every hash in nexts:
            .filter(exists) // Remove undefineds and nulls
            .map(findFromReverseIndex) // Get the Entry from the reverse index
            .reduce(addUniques, []) // Flatten the result and take only uniques
            .concat(nullIndex); // Combine with tails the have no next refs (ie. first-in-their-chain)
        return findUniques(tails, "hash").sort(Entry.compare);
    }
    // Find the hashes to entries that are not in a collection
    // but referenced by other entries
    static findTailHashes(entries) {
        const hashes = {};
        const addToIndex = (e) => (hashes[e.hash] = true);
        const reduceTailHashes = (res, entry, idx, arr) => {
            const addToResult = (e) => {
                /* istanbul ignore else */
                if (hashes[e] === undefined) {
                    res.splice(0, 0, e);
                }
            };
            entry.next.reverse().forEach(addToResult);
            return res;
        };
        entries.forEach(addToIndex);
        return entries.reduce(reduceTailHashes, []);
    }
};
__decorate([
    field({ type: fixedArray("u8", 32) }),
    __metadata("design:type", Uint8Array)
], Log.prototype, "_id", void 0);
Log = Log_1 = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], Log);
export { Log };
//# sourceMappingURL=log.js.map