var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
import { Entry } from "./entry.js";
import { createBlock, getBlockValue } from "@dao-xyz/libp2p-direct-block";
import { BinaryReader, BinaryWriter, deserialize, field, fixedArray, serialize, variant, vec, } from "@dao-xyz/borsh";
import { waitForAsync } from "@dao-xyz/peerbit-time";
import { logger } from "./logger.js";
let Snapshot = class Snapshot {
    id;
    heads;
    size; // we do a size field, because the "true" log size can be larger then the size of values provided below
    values;
    constructor(props) {
        if (props) {
            this.heads = props.heads;
            this.id = props.id;
            this.size = props.size;
            this.values = props.values;
        }
    }
};
__decorate([
    field({ type: fixedArray("u8", 32) }),
    __metadata("design:type", Uint8Array)
], Snapshot.prototype, "id", void 0);
__decorate([
    field({ type: vec("string") }),
    __metadata("design:type", Array)
], Snapshot.prototype, "heads", void 0);
__decorate([
    field({ type: "u64" }),
    __metadata("design:type", BigInt)
], Snapshot.prototype, "size", void 0);
__decorate([
    field({ type: vec(Entry) }),
    __metadata("design:type", Array)
], Snapshot.prototype, "values", void 0);
Snapshot = __decorate([
    variant(0),
    __metadata("design:paramtypes", [Object])
], Snapshot);
export { Snapshot };
export const save = async (snapshotPath, blockstore, cache, log) => {
    const values = await log.getValues();
    const buf = serialize(new Snapshot({
        id: log.id,
        heads: await log.getHeads(),
        size: BigInt(values.length),
        values: values,
    }));
    const snapshot = await blockstore.put(await createBlock(buf, "raw"));
    const writer = new BinaryWriter();
    writer.string(snapshot);
    await cache.set(snapshotPath, writer.finalize());
    await waitForAsync(() => cache.get(snapshotPath).then((bytes) => !!bytes), {
        delayInterval: 200,
        timeout: 10 * 1000,
    });
    logger.debug(`Saved snapshot: ${snapshot}`);
    return snapshot;
};
export const load = async (hash, blockstore) => {
    const block = await blockstore.get(hash);
    if (!block) {
        throw new Error("Missing snapshot for CID: " + hash);
    }
    return deserialize(await getBlockValue(block), Snapshot);
};
export const loadFromCache = async (path, blockstore, cache) => {
    const snapshotOrCID = await cache.get(path);
    if (!snapshotOrCID) {
        throw new Error("Missing snapshot CID from local store");
    }
    const reader = new BinaryReader(snapshotOrCID);
    const snapshotCIDString = reader.string();
    return load(snapshotCIDString, blockstore);
};
//# sourceMappingURL=snapshot.js.map